import typing

import attr
import numpy

from ..components import Component
from ..utils import NRTLParameters, R


class CompositionType:
    """
    A class to describe type of the composition
    """

    molar: str = "molar"
    weight: str = "weight"


@attr.s(auto_attribs=True)
class Mixture:
    """
    A class to represent mixtures
    """

    name: str
    first_component: Component
    second_component: Component
    nrtl_params: NRTLParameters


@attr.s(auto_attribs=True)
class Composition:
    """
    A class to represent composition of the mixtures
    """

    p: float
    type: str

    def __attrs_post_init__(self):
        if self.p < 0:
            self.p = 1e-10
        if self.p > 1:
            self.p = 1-1e-10

    @property
    def first(self) -> float:
        """
        Returns fraction of the first test_components
        """
        return self.p

    @property
    def second(self) -> float:
        """
        Returns fraction of the second test_components
        """
        return 1 - self.p

    def to_molar(self, mixture: Mixture) -> "Composition":
        """
        Converts Composition to molar %
        """
        if self.type == CompositionType.molar:
            return self
        else:
            p = (self.p / mixture.first_component.molecular_weight) / (
                self.p / mixture.first_component.molecular_weight
                + (1 - self.p) / mixture.second_component.molecular_weight
            )
            return Composition(p=p, type=CompositionType.molar)

    def to_weight(self, mixture: Mixture) -> "Composition":
        """
        Converts Composition to weight %
        """
        if self.type == CompositionType.weight:
            return self
        else:
            p = (mixture.first_component.molecular_weight * self.p) / (
                mixture.first_component.molecular_weight * self.p
                + mixture.second_component.molecular_weight * (1 - self.p)
            )
            return Composition(p=p, type=CompositionType.weight)


def get_nrtl_partial_pressures(
    temperature: float, mixture: Mixture, composition: Composition
) -> typing.Tuple[float, float]:
    """
    Calculation of partial pressures of both test_components using NRTL model
    :params
    temperature: temperature in K
    NRTL parameters:
    gij in J/mol (may also be as aij+gij/RT)
    test_mixtures: Mixture
    composition: specified composition in mol or weight %
    :return: Partial pressures as a tuple, test_components wise in kPa
    """
    if composition.type == CompositionType.weight:
        composition = composition.to_molar(mixture=mixture)

    tau = numpy.array(
        [
            (mixture.nrtl_params.a12 + mixture.nrtl_params.g12 / (R * temperature)),
            (mixture.nrtl_params.a21 + mixture.nrtl_params.g21 / (R * temperature)),
        ]
    )
    if mixture.nrtl_params.alpha21 is None:
        alphas = mixture.nrtl_params.alpha12
    else:
        alphas = [mixture.nrtl_params.alpha12, mixture.nrtl_params.alpha21]

    g_exp = numpy.exp(numpy.multiply(-tau, alphas))

    activity_coefficients = [
        numpy.exp(
            (composition.second**2)
            * (
                tau[1]
                * (g_exp[1] / (composition.first + composition.second * g_exp[1])) ** 2
                + tau[0]
                * g_exp[0]
                / (composition.second + composition.first * g_exp[0]) ** 2
            )
        ),
        numpy.exp(
            (composition.first**2)
            * (
                tau[0]
                * (g_exp[0] / (composition.second + composition.first * g_exp[0])) ** 2
                + tau[1]
                * g_exp[1]
                / (composition.first + composition.second * g_exp[1]) ** 2
            )
        ),
    ]

    return (
        mixture.first_component.get_vapor_pressure(temperature)
        * activity_coefficients[0]
        * composition.first,
        mixture.second_component.get_vapor_pressure(temperature)
        * activity_coefficients[1]
        * composition.second,
    )



import typing

import matplotlib.pyplot as plt
import numpy


def plot_graph(
    x_label: str,
    y_label: str,
    points: {str: typing.Tuple[typing.List[float], typing.List[float], bool]},
    title: typing.Optional[str] = "",
):
    """
    Plots a 2D the graph in a pre-defined manner
    :param x_label: The name of the X axis;
    :param y_label: The name of the Y axis;
    :param points:
    if bool is true - line is plotted;
    else - scattered points are plotted;
    :param title: Title of the plot
    :return: shows the plot.
    """

    legend = []

    for key, point_set in points.items():

        if point_set[2]:
            plt.plot(point_set[0], point_set[1])
        else:
            plt.scatter(point_set[0], point_set[1])

        legend.append(key)

    plt.legend(legend)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.suptitle(title)
    plt.show()


def plot_surface(
    condition: bool,
    function: typing.Callable,
    x: typing.List[float],
    t: typing.List[float],
    p: typing.List[float],
    t_min: float,
    t_max: float,
    x_v: numpy.array,
):
    """
    Plots a 3D surface in a pre-defined manner
    :param condition: indicates if experimental points should be scattered
    :param function: Function to be plotted
    :param x: list of experimental values
    :param t: list of experimental values
    :param p: list of experimental values
    :param t_min: minimal value of t parameter to be displayed
    :param t_max: maximal value of t parameter to be displayed
    :param x_v: list of x values to be used for plotting
    :return: plots a 3D graph with or without scattered points
    """
    # TODO: docstring
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")

    if condition:
        ax.scatter(x, t, p, marker="o")

    t_v = numpy.linspace((t_min - 10), (t_max + 10), num=50)
    x_fit, t_fit = numpy.meshgrid(x_v, t_v)
    p_fit = numpy.array([function(x_fit[i], t_fit[i]) for i in range(len(x_fit))])
    ax.plot_surface(x_fit, t_fit, p_fit, alpha=0.2)

    ax.set_xlabel("First component fraction")
    ax.set_ylabel("Temperature K")
    ax.set_zlabel("Permeance")
    fig.suptitle("Fit Illustration", fontsize=10)
    plt.show()



from .components import Component, Components
from .conditions import CalculationType, Conditions, TemperatureProgram
from .diffusion_curve import DiffusionCurve, DiffusionCurveSet
from .experiments import IdealExperiment, IdealExperiments
from .membrane import Membrane
from .mixtures import (
    Composition,
    CompositionType,
    Mixture,
    Mixtures,
    get_nrtl_partial_pressures,
)
from .optimizer import Measurements, PervaporationFunction, find_best_fit, fit
from .permeance import Permeance, Units
from .pervaporation import Pervaporation
from .process import ProcessModel
from .utils import (
    HeatCapacityConstants,
    NRTLParameters,
    R,
    VaporPressureConstants,
    VPConstantsType,
)

__all__ = [
    "VaporPressureConstants",
    "R",
    "NRTLParameters",
    "HeatCapacityConstants",
    "VPConstantsType",
    "ProcessModel",
    "Pervaporation",
    "Permeance",
    "Units",
    "Measurements",
    "PervaporationFunction",
    "find_best_fit",
    "fit",
    "Composition",
    "CompositionType",
    "Mixture",
    "Mixtures",
    "get_nrtl_partial_pressures",
    "Membrane",
    "IdealExperiment",
    "IdealExperiments",
    "DiffusionCurve",
    "DiffusionCurveSet",
    "Conditions",
    "CalculationType",
    "TemperatureProgram",
    "Component",
    "Components",
]

__version__ = "1.1.10"



from .mixture import Composition, CompositionType, Mixture, get_nrtl_partial_pressures
from .mixtures import Mixtures

__all__ = [
    "Mixture",
    "Mixtures",
    "Composition",
    "get_nrtl_partial_pressures",
    "CompositionType",
]



import attr
import numpy

from ..utils import HeatCapacityConstants, R, VaporPressureConstants, VPConstantsType


@attr.s(auto_attribs=True)
class Component:
    name: str
    molecular_weight: float = attr.ib(converter=lambda value: float(value))
    vapour_pressure_constants: VaporPressureConstants
    heat_capacity_constants: HeatCapacityConstants

    def get_vapor_pressure(self, temperature: float) -> float:
        """
        Calculation of saturated pressure in kPa at a given temperature in K using Antoine equation (by the basis of 10)
        type of the calculation using Antoine (antoine) log10(P)=a+b/(T+C)
        or Frost equation (Frost) ln(P) = a+b/T+c/T^2
        :param temperature: temperature in K
        :return: saturated pressure in kPa calculated with respect to constants and given temperature
        """
        if self.vapour_pressure_constants.type == VPConstantsType.antoine:
            return 10 ** (
                self.vapour_pressure_constants.a
                + self.vapour_pressure_constants.b
                / (temperature + self.vapour_pressure_constants.c)
            )
        elif self.vapour_pressure_constants.type == VPConstantsType.frost:
            return numpy.exp(
                self.vapour_pressure_constants.a
                + self.vapour_pressure_constants.b / temperature
                + self.vapour_pressure_constants.c / temperature**2
            )
        else:
            raise ValueError("Type of calculation not supported")

    def get_vaporisation_heat(self, temperature: float) -> float:
        """
        Calculation of Vaporisation heat in kJ/mol using Clapeyron-Clausius equation
        for Antoine equation: H=b*R*ln(10)*(T/(T+C))^2
        for Frost Equation: H=R*(-b-2*c/T)
        :param temperature: temperature in K
        :return: Vaporisation heat in kJ/mol
        """
        if self.vapour_pressure_constants.type == VPConstantsType.antoine:
            return (
                -(
                    (temperature / (temperature + self.vapour_pressure_constants.c))
                    ** 2
                    * R
                    * self.vapour_pressure_constants.b
                    * numpy.log(10)
                )
                / 1000
            )
        elif self.vapour_pressure_constants.type == VPConstantsType.frost:
            return (
                -R
                * (
                    self.vapour_pressure_constants.b
                    + 2 * self.vapour_pressure_constants.c / temperature
                )
                / 1000
            )
        else:
            raise ValueError("Type of calculation not supported")

    def get_specific_heat(self, temperature: float) -> float:
        """
        Calculation of Heat Capacity in J/(mol*K) using polynomial isobaric heat capacity fit
        :param temperature: temperature in K
        :return: Isobaric Heat capacity in J/(mol*K)
        """
        return (
            self.heat_capacity_constants.a
            + self.heat_capacity_constants.b * temperature
            + self.heat_capacity_constants.c * temperature**2
            + self.heat_capacity_constants.d * temperature**3
        )

    def get_cooling_heat(self, t0, t1):
        """
        Calculation of Specific Heat in J/mol using Integral (T2-T1) (CpdT)
        :param t0: temperature in K (t1 < t0)
        :param t1: temperature in K (t1 < t0)
        :return: Specific Heat in J/mol
        """
        return (
            self.heat_capacity_constants.a * (t0 - t1)
            + self.heat_capacity_constants.b * (t0**2 - t1**2) / 2
            + self.heat_capacity_constants.c * (t0**3 - t1**3) / 3
            + self.heat_capacity_constants.d * (t0**4 - t1**4) / 4
        )



from .utils import (
    HeatCapacityConstants,
    NRTLParameters,
    R,
    VaporPressureConstants,
    VPConstantsType,
)

__all__ = [
    "VaporPressureConstants",
    "R",
    "NRTLParameters",
    "HeatCapacityConstants",
    "VPConstantsType",
]



from setuptools import find_packages, setup

setup(
    name="pyvaporation",
    packages=find_packages(),
    version="1.1.10",
    license="Apache license 2.0",
    description="Set of tools for modelling pervaporation processes",
    author="Denis Sapegin, Aleksei Chekmachev",
    author_email="dasapegin@sr-systems.ru",
    url="https://github.com/Membrizard/PyVaporation",
    download_url="https://github.com/Membrizard/PyVaporation/archive/refs/tags/v1.1.5.tar.gz",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    keywords=[
        "pervaporation",
        "membrane",
        "chemistry",
        "modelling",
        "chemical-engineering",
        "scientific"
    ],
    install_requires=[
        "joblib==1.1.0",
        "matplotlib==3.5.2",
        "pandas==1.3.5",
        "scipy==1.7.3",
        "attr==0.3.1",
        "attrs==21.4.0",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



from ..components import Components
from ..utils import NRTLParameters
from .mixture import Mixture


class Mixtures:
    """
    A class with pre-defined mixtures with constants validated against literature data
    For details see ./tests/test_mixtures/test_default_mixtures.py
    """

    H2O_MeOH: Mixture = Mixture(
        name="H2O_MeOH",
        first_component=Components.H2O,
        second_component=Components.MeOH,
        nrtl_params=NRTLParameters(
            g12=-5132.51739,
            g21=1438.40193,
            alpha12=0,
            alpha21=0.3,
            a12=2.7321,
            a21=-0.693,
        ),
    )

    H2O_EtOH: Mixture = Mixture(
        name="H2O_EtOH",
        first_component=Components.H2O,
        second_component=Components.EtOH,
        nrtl_params=NRTLParameters(
            g12=5823,
            g21=-633,
            alpha12=0.3,
        ),
    )

    H2O_iPOH: Mixture = Mixture(
        name="H2O_iPOH",
        first_component=Components.H2O,
        second_component=Components.iPOH,
        nrtl_params=NRTLParameters(
            g12=6899.21,
            g21=106.99,
            alpha12=0.3,
        ),
    )

    H2O_AceticAcid: Mixture = Mixture(
        name="H2O_AceticAcid",
        first_component=Components.H2O,
        second_component=Components.AceticAcid,
        nrtl_params=NRTLParameters(
            g12=-352.42,
            g21=715.43,
            alpha12=0.25,
        ),
    )

    EtOH_ETBE: Mixture = Mixture(
        name="EtOH_ETBE",
        first_component=Components.EtOH,
        second_component=Components.ETBE,
        nrtl_params=NRTLParameters(
            g12=1140.7722,
            g21=2069.17502,
            alpha12=0.3,
        ),
    )

    MeOH_Toluene: Mixture = Mixture(
        name="MeOH_Toluene",
        first_component=Components.MeOH,
        second_component=Components.Toluene,
        nrtl_params=NRTLParameters(
            g12=3857.3,
            g21=4290.3,
            alpha12=0.4370,
        ),
    )

    MeOH_MTBE: Mixture = Mixture(
        name="MeOH_MTBE",
        first_component=Components.MeOH,
        second_component=Components.MTBE,
        nrtl_params=NRTLParameters(
            g12=2133.5,
            g21=2025.3,
            alpha12=0.6,
        ),
    )

    MeOH_DMC: Mixture = Mixture(
        name="MeOH_DMC",
        first_component=Components.MeOH,
        second_component=Components.DMC,
        nrtl_params=NRTLParameters(
            g12=3115.2,
            g21=833.1,
            alpha12=0.3,
        ),
    )
